; A  proof about the mod Tiny program

; Updated to work with stobj, 2/99

(in-package "ACL2")

; Normally we include a book describing Tiny.  But to simplify the
; make process we will assume that it is already loaded

; (include-book "tiny")

(set-verify-guards-eagerness 0) 

(defun program-loaded (st program location)
  (declare (xargs :stobjs (st)))
  (if (consp program)
      (and
       (equal (memi location st) (car program))
       (program-loaded st (cdr program) (1+ location)))
    t))

(defthm mod-prog-lookup
  (implies
   (program-loaded st *mod-prog* 20)
   (and
    (EQUAL (NTH  20 (nth 1 st)) 0)
    (EQUAL (NTH  21 (nth 1 st)) 19)
    (EQUAL (NTH  22 (nth 1 st)) 0)
    (EQUAL (NTH  23 (nth 1 st)) 18)
    (EQUAL (NTH  24 (nth 1 st)) 1)
    (EQUAL (NTH  25 (nth 1 st)) 18)
    (EQUAL (NTH  26 (nth 1 st)) 1)
    (EQUAL (NTH  27 (nth 1 st)) 19)
    (EQUAL (NTH  28 (nth 1 st)) 8)
    (EQUAL (NTH  29 (nth 1 st)) 9)
    (EQUAL (NTH  30 (nth 1 st)) 5)
    (EQUAL (NTH  31 (nth 1 st)) 36)
    (EQUAL (NTH  32 (nth 1 st)) 0)
    (EQUAL (NTH  33 (nth 1 st)) 18)
    (EQUAL (NTH  34 (nth 1 st)) 4)
    (EQUAL (NTH  35 (nth 1 st)) 24)
    (EQUAL (NTH  36 (nth 1 st)) 1)
    (EQUAL (NTH  37 (nth 1 st)) 19)
    (EQUAL (NTH  38 (nth 1 st)) 1)
    (EQUAL (NTH  39 (nth 1 st)) 18)
    (EQUAL (NTH  40 (nth 1 st)) 8)
    (EQUAL (NTH  41 (nth 1 st)) 5)
    (EQUAL (NTH  42 (nth 1 st)) 45)
    (EQUAL (NTH  43 (nth 1 st)) 4)
    (EQUAL (NTH  44 (nth 1 st)) 49)
    (EQUAL (NTH  45 (nth 1 st)) 2)
    (EQUAL (NTH  46 (nth 1 st)) 0)
    (EQUAL (NTH  47 (nth 1 st)) 0)
    (EQUAL (NTH  48 (nth 1 st)) 18)
    (EQUAL (NTH  49 (nth 1 st)) 0)
    (EQUAL (NTH  50 (nth 1 st)) 19)
    (EQUAL (NTH  51 (nth 1 st)) 1)
    (EQUAL (NTH  52 (nth 1 st)) 18)
    (EQUAL (NTH  53 (nth 1 st)) 7))))


(in-theory (disable program-loaded))

(defthm +bv32-simple
  (implies
   (and
    (integerp x) (integerp y)
    (<= x (1- (expt 2 30))) (>= x (- (expt 2 30)))
    (<= y (1- (expt 2 30))) (>= y (- (expt 2 30))))
   (equal (+bv32 x y) (+ x y)))
  :hints (("goal" :in-theory (enable +bv32 signed-byte-p))))

(defthm +bv32-simple2
  (implies
   (and
    (integerp x) (integerp y)
    (<= x (1- (expt 2 31))) (>= x 0)
    (<= y (1- (expt 2 31))) (>= y 0))
   (equal (+bv32 x (- y)) (+ x (- y))))
  :hints (("goal" :in-theory (enable +bv32 signed-byte-p))))
  
(defthm logand-1023-hack
  (implies
   (and
    (<= 0 x)
    (<= x 1023)
    (integerp x))
   (equal (logand x 1023) x))
  :hints (("goal" :in-theory (enable unsigned-byte-p))))

(defthm +bv32-0
  (implies
   (and
    (integerp x)
    (<=  x (1- (expt 2 31)))
    (<= (- (expt 2 31)) x))
   (equal (+bv32 x 0) x))
  :hints (("goal" :in-theory  (enable +bv32 signed-byte-p))))

(defthm tiny-straightline
  (implies
   (syntaxp (quotep n))
   (equal (tiny st n) 
	  (if (zp n) st (tiny (next st) (1- n)))))
  :hints (("goal" :in-theory (disable next))))

(defthm integerp-max
  (implies
   (and (integerp a) (integerp b))
   (integerp (max a b))))

(defthm max-less
  (and
   (equal
    (< (max a b) c)
    (and (< a c) (< b c)))
   (equal
    (< c (max a b))
    (or (< c a) (< c b)))))

(defthm equal-max-x-x
  (implies
   (and (integerp b) (integerp a))
  (and
   (equal (equal (max a b) a) (<= b a))
   (equal (equal (max b a) a) (<= b a))))
  :hints (("goal" :in-theory (enable max))))

(in-theory (disable max))

(defthm equal-plus-x-x
  (implies
   (and (integerp b) (integerp a))
   (equal (equal (+ a b) b) (equal a 0))))

(defthm integerp-unary--
  (implies
   (integerp x)
   (integerp (unary-- x))))

(in-theory (disable tiny))

(defthm less-minus-hack
  (implies
   (and
    (integerp a)
    (integerp b)
    (< 0 a)
    (<= 0 b))
   (not (< a (- b)))))

(defthm lessp-minus-hack3
 (implies
  (and
   (<= a n)
   (<= b n)
   (< 0 b)
   (integerp a)
   (integerp b)
   (integerp n))
  (equal (< n (+ (- b) a)) nil)))

(defthm lessp-minus-hack4
 (implies
  (and
   (<= a n)
   (<= b n)
   (<= 0 b)
   (integerp a)
   (integerp b)
   (integerp n))
  (<= (+ a (- b)) n)))

(defthm max-0-a-b
  (and
   (implies (<= a b) (equal (max 0 (+ a (- b))) 0))
   (implies (>= a b) (equal (max 0 (+ a (- b))) (+ a (- b)))))
  :hints (("goal" :in-theory (enable max))))

(defthm max-open
  (implies
   (and
    (<= a b)
    (integerp a) (integerp b))
   (equal (max a b) b)))

;; We try to backchain first to avoid casesplit.
(defthm nth-set-nth
   (equal
    (nth n1 (set-nth n2 v l))
    (if (equal (nfix n1) (nfix n2))
	v
      (nth n1 l)))
  :hints (("goal" :in-theory (enable set-nth nth))))

(defthm nth-set-nth-1
  (implies
    (not (equal (nfix n1) (nfix n2)))
   (equal
    (nth n1 (set-nth n2 v l))
    (nth n1 l))))

(defthm nth-set-nth-2
  (implies
    (equal (nfix n1) (nfix n2))
   (equal
    (nth n1 (set-nth n2 v l))
    v)))

(defthm nth-cons
  (equal (nth n (cons a b))
	 (if (zp n) a (nth (1- n) b)))
  :hints (("goal" :in-theory (enable nth))))

(defun repeat (n v)
  (if (zp n)
      nil
    (cons v (repeat (1- n) v))))

(defthm set-nth-nil
  (equal (set-nth i v nil)
	 (append (repeat i nil) (list v)))
  :hints (("goal" :in-theory (enable set-nth))))

(defthm set-nth-set-nth-diff
  (implies
   (not (equal (nfix i1) (nfix i2)))
   (equal (set-nth i1 v1 (set-nth i2 v2 l))
	  (set-nth i2 v2 (set-nth i1 v1 l))))
  :hints (("goal" :in-theory (enable set-nth)))
  :rule-classes ((:rewrite :loop-stopper ((i1 i2)))))


(defthm set-nth-set-nth-same
  (equal (set-nth i v1 (set-nth i v2 l))
	 (set-nth i v1 l))
  :hints (("goal" :in-theory (enable set-nth))))

(defthm len-repeat
  (equal (len (repeat i v)) (nfix i)))

(defthm len-set-nth
  (equal (len (set-nth i v l)) (max (1+ (nfix i)) (len l)))
  :hints (("goal" :in-theory (enable set-nth max))))

(defthm car-set-nth
  (equal (car (set-nth i v l)) (if (zp i) v (car l)))
  :hints (("goal" :in-theory (enable set-nth))))  

(defthm len-pushus
  (implies
   (equal (len st) 4)
   (equal (len (pushus val st)) 4)))

(defthm len-mem
  (implies
   (memp1 l n)
   (equal (len l) n)))

(defthm cancel-plus-hack
  (implies
   (and (integerp x) (integerp y))
   (equal (equal (+ y x) y) (equal (fix x) 0))))

(defthm car-append
  (equal (car (append x y))
	 (if (consp x) (car x) (car y))))

(defthm cdr-append
  (equal (cdr (append x y))
	 (if (consp x) (append (cdr x) y) (cdr y))))

(defthm equal-plus-0
  (implies
   (and (integerp a) (integerp b) (<= 0 a) (<= 0 b))
   (equal (equal 0 (+ a b)) (and (equal a 0) (equal b 0)))))

(defthm true-listp-set-nth
  (implies
   (true-listp l)
   (true-listp (set-nth i v l)))
  :hints (("goal" :in-theory (enable set-nth))))

(defthm cancel-hack
  (implies
   (rationalp x)
   (equal (< (+ x y) (+ x z)) (< y z)))
  :rule-classes nil)

(defthm less-quoted-hack
  (implies
   (and
    (rationalp x)
    (syntaxp (quotep a))
    (syntaxp (quotep b))
    (rationalp b))
   (equal (< a (+ b x)) (< (- a b) x)))
  :hints (("goal" :use ((:instance cancel-hack (x (- b)) (y a) (z (+ b x))))
	   :in-theory (disable CANCEL_PLUS-LESSP-CORRECT))))

(defthm less-quoted-hack2
  (implies
   (and
    (rationalp x)
    (syntaxp (quotep a))
    (syntaxp (quotep b))
    (rationalp b))
   (equal (< (+ b x) a) (< x (- a b))))
  :hints (("goal" :use ((:instance cancel-hack (x (- b)) (z a) (y (+ b x))))
	   :in-theory (disable CANCEL_PLUS-LESSP-CORRECT))))

(defthm lessp-hack-of-hacks
  (implies
   (and
    (not (< i1 i2))
    (<= i3 i2))
   (equal (< i1 i3) nil)))

(defthm equal-nat-neg-0
  (implies
   (and
   (integerp n1)
   (integerp n2)
   (<= 0 n1)
   (<= 0 n2))
   (and
    (equal (equal (+ n1 (- n2)) 0) (equal n1 n2))
    (equal (equal (+ (- n2) n1) 0) (equal n1 n2)))))

(defthm car-nthcdr
  (equal (car (nthcdr n l)) (nth n l)))

(in-theory (disable next logand floor len nth-cons set-nth nth nth-cons max))

(in-theory (enable +bv32))

;; From J's library
(defthm iff-equal
   (implies (and (iff x y)
                 (booleanp x)
                 (booleanp y))
            (equal x y))
   :rule-classes nil))

(defthm lessp-unary---hack
  (implies
   (and
    (syntaxp (quotep n2))
    (integerp n1) (integerp n2))
   (equal
    (< (unary-- n1) n2)
    (< (unary-- n2) n1)))
  :hints (("goal" :use (:instance iff-equal (x (< (unary-- n1) n2)) (y (< (unary-- n2) n1))))))
   
(defun mod-loop-once-effect (st)
  (SET-NTH 0 24
	   (SET-NTH 1
		    (SET-NTH 18
			     (+ (NTH 18 (NTH 1 ST))
				(- (NTH 19 (NTH 1 ST))))
			     (SET-NTH (NTH 2 ST)
				      (+ (NTH 18 (NTH 1 ST))
					 (- (NTH 19 (NTH 1 ST))))
				      (SET-NTH (+ -1 (NTH 2 ST))
					       (+ (NTH 18 (NTH 1 ST))
						  (- (NTH 19 (NTH 1 ST))))
					       (NTH 1 ST))))
		    (SET-NTH 2 (NTH 2 ST) ST))))

(defthm mod-loop-repeat
  (implies
   (and
    (stp st)
    (< (dtos st) 1000) (>= (dtos st) 100)
    (equal (progc st) 24)
    (program-loaded st *mod-prog* 20)
    (<= 0 (memi 19 st))
    (<= 0 (memi 18 st))
    (< (memi 19 st) (memi 18 st))
    (not (= (memi 19 st) (memi 18 st)))) ;;redundant
   (equal (tiny st 7)
	  (mod-loop-once-effect st)))
  :hints (("goal''" :in-theory (enable next signed-byte-p))))

; Loop ends, with no adjust needed
(defthm mod-loop-end1
  (implies
   (and
    (stp st)
    (< (dtos st) 1000) (>= (dtos st) 100)
    (<= (ctos st) 1020) (> (ctos st) 1010)
    (equal (progc st) 24)
    (program-loaded st *mod-prog* 20)
    (<= 0 (memi 19 st))
    (< 0 (memi 18 st))
    (> (memi 19 st) (memi 18 st))
    (not (= (memi 19 st) (memi 18 st)))) ;;redundant
   (equal (tiny st 13)
	  (SET-NTH 0
             (LOGAND (NTH (+ 1 (NTH 3 ST)) (NTH 1 ST))
                     1023)
             (SET-NTH 1
                      (SET-NTH 19 0
                               (SET-NTH (NTH 2 ST)
                                        (NTH 18 (NTH 1 ST))
                                        (SET-NTH (+ -1 (NTH 2 ST))
                                                 (+ (- (NTH 18 (NTH 1 ST)))
                                                    (NTH 19 (NTH 1 ST)))
                                                 (SET-NTH (+ -2 (NTH 2 ST))
                                                          (NTH 18 (NTH 1 ST))
                                                          (NTH 1 ST)))))
                      (SET-NTH 2 (+ -1 (NTH 2 ST))
                               (SET-NTH 3 (+ 1 (NTH 3 ST)) ST))))))
  :hints (("goal''" :in-theory (enable next signed-byte-p))))

(defthm mod-loop-end2
  (implies
   (and
    (stp st)
    (< (dtos st) 1000) (>= (dtos st) 100)
    (<= (ctos st) 1020) (> (ctos st) 1010)
    (equal (progc st) 24)
    (program-loaded st *mod-prog* 20)
    (<= 0 (memi 19 st))
    (< 0 (memi 18 st))
    (= (memi 19 st) (memi 18 st)))
   (equal (tiny st 14)
	  (SET-NTH
	   0
	   (LOGAND (NTH (+ 1 (NTH 3 ST)) (NTH 1 ST))
		   1023)
	   (SET-NTH 1
		    (SET-NTH 18 0
			     (SET-NTH 19 0
				      (SET-NTH (NTH 2 ST)
					       0
					       (SET-NTH (+ -1 (NTH 2 ST))
							0
							(SET-NTH (+ -2 (NTH 2 ST))
								 (NTH 18 (NTH 1 ST))
								 (NTH 1 ST))))))
		    (SET-NTH 2 (+ -1 (NTH 2 ST))
			     (SET-NTH 3 (+ 1 (NTH 3 ST)) ST))))))
  :hints (("goal''" :in-theory (enable next signed-byte-p))))

(defun modloop-clock-helper (x y)
  (if (or (not (integerp x)) (not (integerp y)) (>= 0 y) (<= x y))
      (if (equal x y) 14 13)
    (+ 7 (modloop-clock-helper (- x y) y))))

(defun modloop-clock (st)
  (declare (xargs :stobjs (st)))
  (let ((x (memi 18 st)) (y (memi 19 st)))
    (modloop-clock-helper x y)))

(defun remainder (n p)
  (if (or (zp p) (zp n) (< n p))
      (nfix n)
    (remainder (- n p) p)))

(set-irrelevant-formals-ok :warn)

(defun mod-loop-repeat-induct (x y s)
  (if (or (not (integerp x)) (not (integerp y)) (>= 0 y) (<= x y))
      t
    (mod-loop-repeat-induct (- x y) y (mod-loop-once-effect s))))

(defthm less-+-1024-hack
  (implies
   (and
    (integerp n)
    (integerp p))
   (iff (< (+ p n) 1024) (< n (+ (- p) 1024)))))

(defthm tiny-plus
  (implies
   (and
    (integerp x)
    (integerp y)
    (<= 0 x)
    (<= 0 y))
  (equal (tiny s (+ x y))
	 (tiny (tiny s x) y)))
  :hints (("goal" :induct (tiny s x)
	   :in-theory (set-difference-theories (enable tiny) '(next)))))

(defthm memp1-set-nth
  (implies
   (and
    (memp1 l n)
    (integerp a)
    (<= 0 a)
    (< a n))
   (equal (memp1 (set-nth a v l) n)
	  (AND (INTEGERP v) (<= -2147483648 v) (<= v 2147483647))))
  :hints (("goal" :in-theory (enable set-nth))))

(defthm lessp-minus-hack5
  (implies
   (and
    (not (> b (- n)))
    (<= 0 a)
    (integerp a)
    (integerp b)
    (integerp c))
   (not (< (+ a (- b)) n))))

(in-theory (enable max))

(defun sub1-add1-cdr (n1 n2 l)
  (if (consp l)
      (sub1-add1-cdr (1- n1) (1+ n2) (cdr l))
    t))

(defthm set-nth-equal
  (implies
   (and
    (< n (len l))
    (<= 0 n) (integerp n))
   (equal (set-nth n (nth n l) l) l))
  :hints (("goal" :in-theory (enable nth set-nth len))))

(defthm program-loaded-irrelevant
  (implies
   (not (equal n 1))
   (equal (program-loaded (set-nth n v st) prog loc)
	  (program-loaded st prog loc)))
  :hints (("goal" :in-theory (enable set-nth program-loaded))))


;; such a screwy rule.  It's inelegant because program-loaded is a function
;; of state.
(defthm program-loaded-cons
  (implies
   (and
    (or (< n loc) (< (+ (len prog) loc) n) (equal val (nth n mem)))
    (integerp n) (<= 0 n) (integerp loc) (<= 0 loc)
    (program-loaded (set-nth 1 mem st) prog loc))
   (program-loaded (set-nth 1 (set-nth n val mem) st) prog loc))
  :hints (("goal" :in-theory (enable program-loaded set-nth nth len))))

(defthm lessp-hack1
  (implies
   (and
    (<= a b)
    (< c a))
   (equal (< c b) t)))

(defthm set-nth-nth-noop-helper
  (implies
   (and (<= 0 n) (< n (len l)))
   (equal (set-nth n (nth n l) l) l))
  :hints (("goal" :in-theory (enable nth set-nth len))))

(defthm set-nth-nth-noop
  (implies
   (and (<= 0 n) (< n (len l1)) (equal (nth n l1) (nth n l2)))
   (equal (set-nth n (nth n l2) l1) l1)))

(defthm mod-loop-calculates-mod
  (implies
   (and
    (stp st)
    (< (dtos st) 1000) (>= (dtos st) 100)
    (<= (ctos st) 1020) (> (ctos st) 1010)
    (equal (progc st) 24)
    (program-loaded st *mod-prog* 20)
    (equal a (memi 18 st))
    (equal b (memi 19 st))
    (< 0 a) (< 0 b)
    (equal s2 (tiny st (modloop-clock-helper a b))))
   (equal (memi (1+ (dtos s2)) s2) (remainder a b)))
  :rule-classes nil
  :hints (("goal" :in-theory (set-difference-theories
			      (enable stp len)
			      '(tiny-straightline))
			      :induct (mod-loop-repeat-induct a b st))
	  ("Subgoal *1/1.1'" :expand (REMAINDER (NTH 18 (NTH 1 ST))
                       (NTH 19 (NTH 1 ST))))))

(defun remclock (st)
  (declare (xargs :stobjs (st)))
  (modloop-clock-helper (memi 18 st) (memi 19 st)))

(defun good-initial-remainder-state (st)
  (declare (xargs :stobjs (st)))
   (and
    (stp st)
    (< (dtos st) 1000) (>= (dtos st) 100)
    (<= (ctos st) 1020) (> (ctos st) 1010)
    (equal (progc st) 24)
    (program-loaded st *mod-prog* 20)
    (< 0 (memi 18 st))
    (< 0 (memi 19 st))))

(defthm remainder-correct
  (let ((s2 (tiny st (remclock st))))
  (implies
    (good-initial-remainder-state st)
    (equal (memi (1+ (dtos s2)) s2)
           (remainder (memi 18 st) (memi 19 st)))))
  :rule-classes nil
  :hints (("goal"
	   :use ((:instance mod-loop-calculates-mod
			    (s2 (tiny st (modloop-clock-helper (memi 18 st) (memi 19 st))))
			    (a (memi 18 st))
			    (b (memi 19 st))))
	   :in-theory (disable tiny-straightline))))
